"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.addRxPlugin = addRxPlugin;

var _rxSchema = require("./rx-schema");

var _crypter = require("./crypter");

var _rxDocument = require("./rx-document");

var _rxQuery = require("./rx-query");

var _rxCollection = require("./rx-collection");

var _rxDatabase = require("./rx-database");

var _overwritable = require("./overwritable");

var _hooks = require("./hooks");

var _rxError = require("./rx-error");

/**
 * this handles how plugins are added to rxdb
 * basically it changes the internal prototypes
 * by passing them to the plugins-functions
 */

/**
 * prototypes that can be manipulated with a plugin
 */
var PROTOTYPES = {
  RxSchema: _rxSchema.RxSchema.prototype,
  Crypter: _crypter.Crypter.prototype,
  RxDocument: _rxDocument.basePrototype,
  RxQuery: _rxQuery.RxQueryBase.prototype,
  RxCollection: _rxCollection.RxCollectionBase.prototype,
  RxDatabase: _rxDatabase.RxDatabaseBase.prototype
};
var ADDED_PLUGINS = new Set();
/**
 * Add a plugin to the RxDB library.
 * Plugins are added globally and cannot be removed.
 */

function addRxPlugin(plugin) {
  (0, _hooks.runPluginHooks)('preAddRxPlugin', {
    plugin: plugin,
    plugins: ADDED_PLUGINS
  }); // do nothing if added before

  if (ADDED_PLUGINS.has(plugin)) {
    return;
  } else {
    ADDED_PLUGINS.add(plugin);
  }
  /**
   * Since version 10.0.0 we decoupled pouchdb from
   * the rxdb core. Therefore pouchdb plugins must be added
   * with the addPouchPlugin() method of the pouchdb plugin.
   */


  if (!plugin.rxdb) {
    throw (0, _rxError.newRxTypeError)('PL1', {
      plugin: plugin
    });
  } // prototype-overwrites


  if (plugin.prototypes) {
    Object.entries(plugin.prototypes).forEach(function (_ref) {
      var name = _ref[0],
          fun = _ref[1];
      return fun(PROTOTYPES[name]);
    });
  } // overwritable-overwrites


  if (plugin.overwritable) {
    Object.assign(_overwritable.overwritable, plugin.overwritable);
  } // extend-hooks


  if (plugin.hooks) {
    Object.entries(plugin.hooks).forEach(function (_ref2) {
      var name = _ref2[0],
          fun = _ref2[1];
      return _hooks.HOOKS[name].push(fun);
    });
  }
}

//# sourceMappingURL=plugin.js.map